/* 
 *    tpuart -- A TP-UART driver 
 *
 *
 *    File: transmit.c -- transmit routine
 *
 *    $Id: transmit.c,v 1.2 2001/05/23 11:33:49 rst Exp $
 *
 *
 *    Copyright (C) 2000 Raffael Stocker <raffael.stocker@stud.fh-deggendorf.de> 
 *                       University of applied sciences Deggendorf
 *
 *    This program is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU General Public License as
 *    published by the Free Software Foundation; either version 2 of the
 *    License, or (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *    General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place--Suite 330, Boston, MA
 *    02111-1307, USA.
 *
 *
 *    07-Aug-2000: 0.01  Started Coding
 *
 *    Last change:
 *
 *    $Date: 2001/05/23 11:33:49 $ $Author: rst $
 */

#include <linux/config.h>
#ifdef CONFIG_MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/kernel.h>
#include <linux/sched.h>
#include <asm/io.h>
#include "tpuart_main.h"

void transmit (void *arg)
{
	tpuart_t *dev = arg;
	unsigned char padbyte = U_L_DATA_START;
	unsigned short length;
	unsigned short pos = 0;
	unsigned long flags;
	
	if (test_bit (LET_ME_ALONE_I_AM_ALREADY_SENDING, &dev->tx.flags)) {
		/* if we can't send now, signal sending request */
		PDEBUG ("Hell, can't send...\n");
		set_bit (SEND_REQU, &dev->tx.flags);
	} else {
		/* sending in int_handler must be complete at this point! */
		dev->tx.write = dev->tx.buffer;
     
		if (!deque_get_irq (&dev->tx.temp, &dev->tx.tel_buffer)) {
			/* appearently nothing to read */
			PDEBUG ("Tried to read empty queue!\n");
			return;
		}
		PDEBUG ("Sending message!\n");
		clear_bit (SEND_REQU, &dev->tx.flags);
		set_bit (LET_ME_ALONE_I_AM_ALREADY_SENDING, &dev->tx.flags);
		set_bit (DID_SEND, &dev->tx.flags);

		length = dev->tx.temp.data_length;
		if (length > 64) /* sanity check */
			length = 64;

		for (; length > 1; length--) {
			PDEBUG ("Queuing padbyte: 0x%x msgbyte: 0x%x\n", padbyte, 
				dev->tx.temp.data[pos]);
			*dev->tx.write++ = padbyte++;
			*dev->tx.write++ = dev->tx.temp.data[pos++];
		}
		PDEBUG ("Sending last padbyte: 0x%x msgbyte: 0x%x\n", (padbyte | 0x40) & 0x7F,
			dev->tx.temp.data[pos]);
		*dev->tx.write++ = (padbyte | 0x40) & 0x7F;  /* end code is 0x4i (i = index) */
		*dev->tx.write = dev->tx.temp.data[pos];
		dev->tx.end = dev->tx.write;
		dev->tx.read = dev->tx.buffer;
		/* write out first 2 bytes to provocate tx interrupt */
		set_bit (SEND_DATA, &dev->tx.flags);
		/*clear_bit (GOT_TP_ERROR, &dev->tx.flags);*/
		save_flags (flags); cli ();
		outb (*dev->tx.read++, dev->port);
		outb (*dev->tx.read++, dev->port);
		restore_flags (flags);
	}
}
